# Multiproject Plugin Specification

**Owners:**
Pepper Lebeck-Jobe [@eljobe](https://github.com/eljobe)
Stefan Oehme [@oehme](https://github.com/oehme)   
**Updated:** 2016-07-07   
![Approved](https://img.shields.io/badge/design-approved-green.svg)

<!-- START doctoc generated TOC please keep comment here to allow auto update -->
<!-- DON'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->


- [Objective](#objective)
  - [Goals / Use Cases](#goals--use-cases)
  - [Non Goals / Out of Scope Use Cases](#non-goals--out-of-scope-use-cases)
- [Background](#background)
- [Solution](#solution)
- [Test Plan](#test-plan)

<!-- END doctoc generated TOC please keep comment here to allow auto update -->

## Objective

Allow users to apply plugins to multiple projects at once using the new `plugins` DSL.

### Goals / Use Cases

* As a build author, I want to apply a plugin to all projects in my build. For instance, all projects in my build should apply the "idea" plugin.
* As a build author, I want to apply a plugin to all sub-projects below the current one. For instance, all sub-projects of the root are "java" projects.
* As a build author, I want to conditionally apply a plugin, e.g. based on the name of a project, one of its properties or some environment variable
* As a build author, I want to [cross-configure](https://docs.gradle.org/current/userguide/multi_project_builds.html#sec:cross_project_configuration) the conventions and tasks of plugins that I apply. To that end, their classes need to be on the parent’s classpath, even if the parent does not apply the plugin to itself.

### Non Goals / Out of Scope Use Cases

* As a build author, I want to apply plugins to the projects in a build from an init-script (e.g. part of a custom Gradle distribution).
* As a Tooling API integrator, I want to retrieve which plugins are applied to a project and which classes are visible to it’s build script. Retrieving that information must be fast, so I can provide timely auto completion proposals to the IDE user.
* As a build user, I want to know which plugins were applied to a project and where they were applied from.
* As a build author, I want to be warned if incompatible plugin requests are made. E.g. a parent project applying version "1" of a plugin and a child project trying to apply version “2”.

## Background

The new, concise and declarative `plugins` block has a major disadvantage that makes it hard to use in multi-project builds: Each project must declare its plugins itself, they currently cannot be cross-configured. This leads to a lot of repetition. As a result, the `plugins` block is mostly used in single-project builds.

With the old `buildscript` block you can put a plugin on the parent project’s classpath and then apply it to all children like this:

```gradle
buildscript {
  repositories.jcenter()
  dependencies {
    classpath 'org.foo:my-plugin:1.0'
  }
}

subprojects {
  apply plugin: 'my-plugin'
}
```

You also have access to many of the subproject’s properties at this point, like its name, location and properties defined in the `gradle.properties` file.

The new `plugins` block needs a similar flexibility so it can replace the old mechanism, but at the same time needs to prevent the user from causing unwanted side effects. Evaluating this block quickly and at any time we want will be essential e.g. for IDE-support (the IDE needs to know which plugins are applied to give sensible completion proposals).

## Solution

Based on the feedback we collected in the [Cross Configuration Deep Thinking](https://docs.google.com/document/d/18kMjtZ5xhGmmEIpRSFX3AM1msflz-lu7MKViL83pLAs/edit#heading=h.3mmc8cemoacc) document, we realized that a lazy, parallelizable and flexible cross-configuration mechanism will be needed in the future, but is out of scope for this project.

However, build scripts will definitely need a convenient syntax to declare a dependency on another plugin. This also solves the current problem of making a plugin available without applying it to the parent project.

We add an a new `apply` function to the plugins block which can be set to `false` to allow users to resolve a plugin and put its classes on the buildscript’s classpath without applying it. Users can then use the usual `apply plugin` syntax to apply this resolved plugin to subprojects.

```gradle
plugins {
  id "nebula.source-jar" version "4.4.4" apply false
}

subprojects {
  if (openSource) {
    apply plugin: "nebula.source-jar"
  }
}
```

The classloading semantics stay unchanged, i.e. subprojects still see all classes loaded from the parent project and all plugins are put into a flat classloader to which the buildscript has unconstrained access.

## Test Plan

* plugins block can contain the new `apply` function
* `apply` is mentioned in error messages as a valid method
* When `apply false` is specified for a plugin, it is resolved, but not applied to the current project
* After resolving a plugin, the current project’s build script has access to classes contributed by that plugin
* After resolving a plugin, the subproject’s build scripts have access to classes contributed by that plugin
